//
// file :         readthread.cpp
//
// description :  ramping thread
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AREA Science Park
//              34012 Basovizza, Trieste ITALY
//

#include "readthread.h"
#include "CCDSpectrometer.h"
#include <math.h>
#include <iomanip>	//used for debug dump of db
#ifdef HAVE_RNM
#include "k_rnmshare.h"
#endif

namespace CCDSpectrometer_ns
{

//+------------------------------------------------------------------
//
//	method:			readthread::readthread()
//
//	description:	readthread constructor
//
//-------------------------------------------------------------------
//readthread::readthread(Tango::DeviceImpl* devImpl):Tango::LogAdapter(devImpl),cond(&mut)
readthread::readthread(Tango::DeviceImpl* devImpl):Tango::LogAdapter(devImpl)
{ 
	DEBUG_STREAM << "readthread::readthread(): constructor..." << endl;
	device = devImpl;
	waitflag = true;
	last_camera_counter = 0;
#ifdef HAVE_RNM
	int ret = rnmshare_init(0);
#endif
}

//+------------------------------------------------------------------
//
//	method:			readthread::~readthread()
//
//	description:	readthread destructor
//
//-------------------------------------------------------------------
readthread::~readthread()
{
	DEBUG_STREAM << "readthread::~readthread(): destructor... !" << endl;
#ifdef HAVE_RNM
	rnmshare_deinit();
#endif
}
#if 0
union
{
	uint16_t value;
	uint8_t data[2];
} chat_to_int16;
#endif


string dump_hex(const vector<unsigned char> &data)
{
	stringstream mesg_hex;

	for (unsigned int i=0; i<data.size(); ++i) {
		mesg_hex << setw(2) << setfill('0')
			<< hex << (int)data[i] << " ";
	}

	return mesg_hex.str();
}


//+------------------------------------------------------------------
//
//	method:			readthread::run()
//
//	description:	Run
//
//-------------------------------------------------------------------
void readthread::run(void *)
{
	INFO_STREAM << "readthread::run(): running... !" << endl;
	/*int pausesec,pausenano,pausesec_dfl,pausenano_dfl,
	    cnt_err = 0,
		max_cnt_err = 3,
		ret;*/
	pausesec = 1;
	pausenano = 0;
	//mutex_sleep->unlock();
	double *pNoiseData=NULL;
	int flag;

	unsigned long counter =0;
	abortflag = false;
	CCDSpectrometer *_device = static_cast<CCDSpectrometer *>(device);

	INFO_STREAM << "readthread::run(): before while(!abortflag... !" << endl;
	while (!abortflag/* && !_device->abortflag*/)
	{
		DEBUG_STREAM << "readthread::run(): pausasec="<<pausesec<<" init_live_mode="<<(int)_device->init_live_mode<< " live_mode="<<_device->live_mode<<" abortflag="<<(int)abortflag<<endl;
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++INIT MODE+++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		if(_device->init_live_mode || _device->init_measure_mode || _device->init_acquire_noise_mode)
		{
			{
				Tango::DevLong num_pixel = _device->numPixel;//1280;//TODO: read num pixel!!
				//_device->dp->read_attribute()>>num_pixel;//TODO: read num pixel!!
				_device->numPixel = num_pixel;//TODO: read num pixel!!

				DEBUG_STREAM << "readthread::run() allocating for Data size=" << ((int)_device->numPixel)/**((int)_device->m_usbCCD.m_nNumReadOuts)*/ << endl;
				if((!_device->init_acquire_noise_mode) && _device->pLiveData)
				{
					delete [] _device->pLiveData;
					_device->pLiveData = NULL;
				}
				if(!_device->init_acquire_noise_mode)
				{
					_device->pLiveData = new double[((int)_device->numPixel)/**((int)_device->m_usbCCD.m_nNumReadOuts)*/];
					memset(_device->pLiveData,0,((int)_device->numPixel)*sizeof(double));
				}
				else
				{
					pNoiseData = new double[((int)_device->numPixel)/**((int)_device->m_usbCCD.m_nNumReadOuts)*/ * _device->noise_frames];
					memset(pNoiseData,0,((int)_device->numPixel)* _device->noise_frames*sizeof(double));

				}
				//DEBUG_STREAM << "readthread::run() allocated for Data size=" << ((int)_device->numPixel)*((int)_device->m_usbCCD.m_nNumReadOuts) << endl;
				/*if(_device->nFramesBuffered > 0)
				{
					_device->pFrameBuffer = new DWORD[_device->numPixel * _device->nFramesBuffered];
					memset(_device->pFrameBuffer,0,_device->numPixel * _device->nFramesBuffered * sizeof(DWORD));
				}*/
				//mutex_sleep->lock();
				pausesec = 0;
				//pausenano = 1000000;		//TODO: 1 msec, it should depend on freq of CCD
				//pausenano = (*(_device->attr_ExposureTime_read) * *(_device->attr_NumReadouts_read) * 300000);		//TODO: 10 msec, it should depend on freq of CCD
				pausenano = 0;
				DEBUG_STREAM << "readthread::run(): setting sleep_time(ns)="<<pausenano<<endl;
				//mutex_sleep->unlock();
				_device->init_live_mode = false;
				_device->init_measure_mode = false;
				_device->init_acquire_noise_mode = false;
				counter =0;
			}
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++ACQUIRE MODE++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		if(_device->live_mode || _device->measure_mode || _device->stop_mode)
		{
			//DWORD stat;
			//DEBUG_STREAM << "readthread::run(): before GetMeasStatus"<<endl;
			//_device->m_usbCCD.GetMeasStatus(&stat);
			//if(stat == STATUS_DATA_READY)

				//if(_device->m_usbCCD.m_nNumReadOuts == 1)
				if(1)
				{
					if(*(_device->attr_NoiseSubtraction_read))
					{
						//DEBUG_STREAM << "readthread::run(): before GetData readouts=1, noise subtraction"<<endl;
						if(counter==0 || (*(_device->attr_SyncMode_read) != 0))
							flag = PrepareMeas();
						else
							flag = 0;
						if(flag==0)
							flag = Measure(_device->pLiveData);
						if(flag < 0)
						{
							if(flag==-1)
							{
								_device->set_state(Tango::FAULT);
								//_device->set_status("Communication error");
								//_device->disconnect_device();
								/*if(_device->live_mode)
									_device->init_live_mode = true;
								if(_device->measure_mode)
									_device->init_measure_mode = true;*/
								_device->live_mode = false;
								_device->init_live_mode = false;
								_device->measure_mode = false;
								_device->init_measure_mode = false;
								_device->stop_mode = false;
								pausesec = 1;
								pausenano = 0;
								_device->conn_status = false;
							}
							else if(flag==-2) 
							{
								_device->set_state(Tango::OFF);
							}
							continue;
						}
						else if(flag > 0)
						{
							if(_device->stop_mode)
							{
								_device->live_mode=false;
								_device->measure_mode=false;
							}
							else
							{
								_device->set_state(Tango::ALARM);
							}
							//_device->set_status("Trigger Missing");
							continue;
						}
						else
						{
							if(_device->get_state()==Tango::ALARM || _device->get_state()==Tango::OFF)
							{
								_device->set_state(Tango::RUNNING);
								_device->set_status("Continuous Mode");
							}
						}
						for(unsigned int j=0; j < _device->numPixel; j++)
						{
							//_device->pLiveData[j] = (((int)_device->pLiveData[j] - _device->noise_level) > 0) ? ((int)_device->pLiveData[j] - _device->noise_level) : 0;
							if(*_device->attr_IntensityCalibrationEnable_read)
							{
								_device->pLiveData_avg[j] = (_device->pLiveData[j] - _device->pNoiseMask_avg[j]);
								if(_device->pLiveData_avg[j] > 0)
									_device->pLiveData_avg[j] *= _device->attr_IntensityCalibration_read[j];
							}
							else
							{
								_device->pLiveData_avg[j] = (_device->pLiveData[j] - _device->pNoiseMask_avg[j]);
							}
							/*//TTTTTTTTTTTTTTTTEEEEEEEEEEEEEEEEESSSSSSSSSSSSSSTTTTTTTTTTTTTTTTTTTTTTTTTTTT: generate fake gaussian
							double base = 0 + (-1 + 2*(counter%2)) * counter%101;
							double amp =15000 + (1 - 2*(counter%2)) * counter%547;
							double center = 350 + (-1 + 2*(counter%2)) * counter%3;
							double width = 10  + (1 - 2*(counter%2)) * counter%2;
							_device->pLiveData_avg[j] += base + amp*exp((-0.5) * pow((_device->x_values[j]-center)/width,2));*/
						}
						//memcpy(_device->pLiveData_avg,_device->pLiveData,_device->numPixel * sizeof(DWORD));
					}
					else
					{
						//DEBUG_STREAM << "readthread::run(): before GetData readouts=1"<<endl;
						if(counter==0 || (*(_device->attr_SyncMode_read) != 0))
							flag = PrepareMeas();
						else
							flag = 0;
						if(flag==0)
							flag = Measure(_device->pLiveData);
						if(flag < 0)
						{
							if(flag==-1)
							{
								_device->set_state(Tango::FAULT);
								//_device->set_status("Communication error");
								//_device->disconnect_device();
								/*if(_device->live_mode)
									_device->init_live_mode = true;
								if(_device->measure_mode)
									_device->init_measure_mode = true;*/
								_device->live_mode = false;
								_device->init_live_mode = false;
								_device->measure_mode = false;
								_device->init_measure_mode = false;
								_device->stop_mode = false;
								pausesec = 1;
								pausenano = 0;
								_device->conn_status = false;
							}
							else if(flag==-2) 
							{
								_device->set_state(Tango::OFF);
							}
							continue;
						}
						else if(flag > 0)
						{
							if(_device->stop_mode)
							{
								_device->live_mode=false;
								_device->measure_mode=false;
							}
							else
							{
								_device->set_state(Tango::ALARM);
							}
							//_device->set_status("Trigger Missing");
							continue;
						}
						else
						{
							if(_device->get_state()==Tango::ALARM || _device->get_state()==Tango::OFF)
							{
								_device->set_state(Tango::RUNNING);
								_device->set_status("Continuous Mode");
							}
						}
						if(*_device->attr_IntensityCalibrationEnable_read)
						{
							for(unsigned int j=0; j < _device->numPixel; j++)
								_device->pLiveData_avg[j] = (_device->attr_IntensityCalibration_read[j] * _device->pLiveData[j]);
						}
						else
						{
							memcpy(_device->pLiveData_avg,_device->pLiveData,_device->numPixel * sizeof(double));
						}
					}
					/*for(unsigned int i=0; i<_device->numPixel; i++)
					{
						DEBUG_STREAM << "OceanOptics::OceanOptics(): _device->pLiveData["<<i<<"]=" << _device->pLiveData[i] << endl;
					}*/
					DEBUG_STREAM << __func__<<": READ DATA" << endl;
					clock_gettime(CLOCK_MONOTONIC,&_device->last_read);
				}
				/*else
				{
					//DEBUG_STREAM << "readthread::run(): before GetData readouts>1"<<endl;
					_device->m_usbCCD.GetData(_device->pLiveData);
					//memset(&(_device->pLiveData_avg[0]), 0,_device->numPixel*sizeof(DWORD));

					//for every pixel
					for(unsigned int j=0; j < _device->numPixel; j++)
					{
						_device->pLiveData_avg[j] = 0;
						//sum value in this pixel point for every frame
						for(int i=0; i < _device->m_usbCCD.m_nNumReadOuts; i++)
						{
							_device->pLiveData_avg[j] += _device->pLiveData[i * _device->numPixel + j];
						}
						//divide the sum per the number of frames
						_device->pLiveData_avg[j] = (DWORD)((((double)_device->pLiveData_avg[j])/((double)_device->m_usbCCD.m_nNumReadOuts)) + 0.5); //+0.5 to round and not truncate
						if(*(_device->attr_NoiseSubtraction_read))
						{
							//_device->pLiveData_avg[j] = (((int)_device->pLiveData_avg[j] - _device->noise_level) > 0) ? ((int)_device->pLiveData_avg[j] - _device->noise_level) : 0;
							_device->pLiveData_avg[j] = (int)_device->pLiveData_avg[j] - _device->noise_level;
						}
					}
				}*/
				if(_device->nFramesBuffered > 0)
				{
					//copy the average frame
					memcpy(&(_device->pFrameBuffer[_device->numPixel * _device->FrameBufferIndex]), &(_device->pLiveData_avg[0]),_device->numPixel*sizeof(double));
					////update index in buffer once for all readout frames
					_device->FrameBufferIndex = (_device->FrameBufferIndex + 1) % _device->nFramesBuffered;
				}

				struct timeval buf_time;
				gettimeofday(&buf_time,NULL);
				int32_t buf_bunch_num;


				if(*_device->attr_RtMode_read)
				{
#ifdef HAVE_RNM
					char check;
					rnm_bunch_number_get_value(0, 1, &buf_bunch_num, &check);
					//buf_bunch_num = fit_counter;
#endif
					if(old_bunch_number != buf_bunch_num-1)
					{
						//INFO_STREAM << "readthread::run(): BUNCH error: previous="<<old_bunch_number<<" present="<<bunch_number;
						//cout << "--------------------------------------->>>>>>>>>>>>>>>>>>>>>>>>> BUNCH error: previous="<<old_bunch_number<<" present="<<bunch_number<<"<<<<<<<<<<<<<<<<<<<<<<<<<---------------------------"<<endl;
						INFO_STREAM << "--->>> BUNCH error: previous="<<old_bunch_number<<" present="<<buf_bunch_num<<"<<<---"<<endl;
						old_bunch_number = buf_bunch_num;
					}
					else
					{
						old_bunch_number++;
					}
				}
				else
				{
					buf_bunch_num = fit_counter;
				}
				//cout << __func__<<": before push_data data[0]="<<_device->pLiveData_avg[0]<<" data[1]="<<_device->pLiveData_avg[1]<<" data[2]="<<_device->pLiveData_avg[2]<<" data[1000]="<<_device->pLiveData_avg[1000]<<" data[1001]="<<_device->pLiveData_avg[1001]<<" data[1002]="<<_device->pLiveData_avg[1002]<<endl;
				tbuffer::_data_array<double> val_array(_device->pLiveData_avg, _device->numPixel);
				_device->tbuf_array.push_data(&val_array,buf_time,buf_bunch_num);
				tbuffer::_data_array<double> val_array_w(&(_device->pLiveData_avg[0 + _device->RegionOfInterestStartPix]), _device->RegionOfInterestEndPix - _device->RegionOfInterestStartPix + 1);
				_device->tbuf_array_w.push_data(&val_array_w,buf_time,buf_bunch_num);
				tbuffer::_data_array<double> val_array_l(&(_device->x_values[0 + _device->RegionOfInterestStartPix]), _device->RegionOfInterestEndPix - _device->RegionOfInterestStartPix + 1);
				_device->tbuf_array_l.push_data(&val_array_l,buf_time,buf_bunch_num);
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::::::::::  FWHM processing :::::::::::::::::::::::
				//________________________________________________________________
				if(*(_device->attr_FWHMEnable_read))
				{
					size_t x_peak, x_half_min, x_half_max;
					double y_peak;
					_device->fwhm_calculate(_device->pLiveData_avg,_device->numPixel,
						_device->AnalysisRegionStartPix,_device->AnalysisRegionEndPix,
						1,				//average on 1 frame
						*(_device->attr_FWHMAvg_read),				//average between 3 adjacent values
						x_peak, y_peak,
						x_half_min, x_half_max);
					_device->attr_FWHM_read[0]=_device->x_values[x_peak];
					_device->attr_FWHM_read[1]=y_peak;
					if(*(_device->attr_WidthCorrectionEnable_read))
					{
						_device->attr_FWHM_read[2]=sqrt((_device->x_values[x_half_max] - _device->x_values[x_half_min])*(_device->x_values[x_half_max] - _device->x_values[x_half_min]) + _device->widthCorrection*_device->widthCorrection);
					}
					else
					{
						_device->attr_FWHM_read[2]=_device->x_values[x_half_max] - _device->x_values[x_half_min];
					}
					*(_device->attr_FWHMCenter_read) = _device->attr_FWHM_read[0];
					*(_device->attr_FWHMWidth_read) = _device->attr_FWHM_read[2];
					double val_buff;
					val_buff = *(_device->attr_FWHMCenter_read);
					_device->tbuf[BUF_ID_FWHM_CENTER].push_data(&val_buff,buf_time,buf_bunch_num);
					val_buff = *(_device->attr_FWHMWidth_read);
					_device->tbuf[BUF_ID_FWHM_WIDTH].push_data(&val_buff,buf_time,buf_bunch_num);
					if(!(*(_device->attr_GaussFitEnable_read)) || _device->preferFWHM)
						*(_device->attr_Center_read) = _device->attr_FWHM_read[0];
						*(_device->attr_Width_read) = _device->attr_FWHM_read[2];
				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//::::::::::::::::::::::::  AUTO EXPOSURE ::::::::::::::::::::::::
				//________________________________________________________________
				if(*(_device->attr_AutoExposure_read))
				{
					double y_max;
					if(*(_device->attr_FWHMEnable_read))
					{
						y_max=_device->attr_FWHM_read[1];
					}
					else
					{
						size_t x_peak, x_half_min, x_half_max;
						double y_peak;
						_device->fwhm_calculate(_device->pLiveData_avg,_device->numPixel,
							_device->AnalysisRegionStartPix,_device->AnalysisRegionEndPix,
							1,				//average on 1 frame
							*(_device->attr_FWHMAvg_read),				//average between 3 adjacent values
							x_peak, y_peak,
							x_half_min, x_half_max);
						y_max = y_peak;
					}
					_device->exposures[_device->exposure_idx] = y_max;
					_device->exposure_idx = (_device->exposure_idx + 1)%AUTO_EXPOSURE_MEAN;
					_device->y_max_mean = 0;
					for(int i=0; i<AUTO_EXPOSURE_MEAN; i++)
						_device->y_max_mean += _device->exposures[i];
					_device->y_max_mean = (int)(((double)_device->y_max_mean)/AUTO_EXPOSURE_MEAN);

					DEBUG_STREAM << "readthread::run(): AutoExposure: max="<<_device->y_max_mean<<endl;

					if(_device->y_max_mean > _device->autoExposureAmpRange[1])
					{
						double exposure = *(_device->attr_ExposureTime_read);
						if(exposure - _device->autoExposureStep >= _device->autoExposureRange[0])
						{
							exposure -= _device->autoExposureStep;
						}
						else
						{
							exposure = _device->autoExposureRange[0];
						}
						if(exposure != *(_device->attr_ExposureTime_read))
						{
							DEBUG_STREAM << "readthread::run(): AutoExposure: decreasing to "<<exposure<<endl;
						    *(_device->attr_ExposureTime_read) = exposure;
						}
					}
					else if(_device->y_max_mean < _device->autoExposureAmpRange[0])
					{
						double exposure = *(_device->attr_ExposureTime_read);
						if(exposure + _device->autoExposureStep <= _device->autoExposureRange[1])
						{
							exposure += _device->autoExposureStep;
						}
						else
						{
							exposure = _device->autoExposureRange[1];
						}
						if(exposure != *(_device->attr_ExposureTime_read))
						{
							DEBUG_STREAM << "readthread::run(): AutoExposure: increasing to "<<exposure<<endl;
						    *(_device->attr_ExposureTime_read) = exposure;
						}
					}
				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::::::::: Gaussian Fitting ::::::::::::::::::::::::
				//________________________________________________________________
				if(*(_device->attr_GaussFitEnable_read))
				{

					int ret_fun;
					double fit[CONFITEOR_FIT_NUM_PARAM];
					//if last fit ok, initial guess is last fit
					if(_device->GaussFit_valid || use_last_fit)
					{
						fit[0] = _device->attr_GaussFit_read[0];
						fit[1] = (_device->attr_GaussFit_read[1] > 0) ? _device->attr_GaussFit_read[1] :
							((*(_device->attr_FWHMEnable_read) && (_device->attr_FWHM_read[1] > 0)) ?
							_device->attr_FWHM_read[1] : 50 + (10*counter % 950));
						int center_ind=(int)(((double)(_device->AnalysisRegionEndPix - _device->AnalysisRegionStartPix + 1))/2);
						//if center is inside analysis region use it, otherwise use fwhm, otherwise use center of analysis region
						fit[2] = (_device->attr_GaussFit_read[2] > _device->x_values[_device->AnalysisRegionStartPix] && _device->attr_GaussFit_read[2] < _device->x_values[_device->AnalysisRegionEndPix]) ?
							_device->attr_GaussFit_read[2] : ((*(_device->attr_FWHMEnable_read) && (_device->attr_FWHM_read[0] > 0)) ? _device->attr_FWHM_read[0] : _device->x_values[center_ind]);
						double width = (_device->x_values[_device->AnalysisRegionEndPix] - _device->x_values[_device->AnalysisRegionStartPix]);

						double width_fit_uncalibr;
						if(*(_device->attr_WidthCorrectionEnable_read))
						{
							width_fit_uncalibr=sqrt(_device->attr_GaussFit_read[3]*_device->attr_GaussFit_read[3] - _device->widthCorrection*_device->widthCorrection);
						}
						else
						{
							width_fit_uncalibr=_device->attr_GaussFit_read[3];
						}

						//if rms computed is too big, use last fwhm, otherwise use a fraction of analysis region
						fit[3] = (width_fit_uncalibr > 0 && width_fit_uncalibr < width) ? width_fit_uncalibr*SIGMA_TO_FWHM_FACTOR :
							((*(_device->attr_FWHMEnable_read) && (_device->attr_FWHM_read[2] > 0)) ? _device->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR : width/10.0);
						fit[4] = fit[3];
						DEBUG_STREAM << __func__<<": initial guess with GaussFit_valid="<<(int)_device->GaussFit_valid<<" use_last_fit="<<(int)use_last_fit<<" fit[0]="<<fit[0]<<" fit[1]="<<fit[1]<<" fit[2]="<<fit[2]<<" fit[3]="<<fit[3]<<endl;
					}
					//otherwise if center and width reference are defined, use that (but not always, maybe they are set to bad values)
					else if ((_device->attr_GaussReference_read[2]) != 0 && (_device->attr_GaussReference_read[3]) && (counter % 10 == 0))
					{
						fit[0] = _device->attr_GaussReference_read[0];
						fit[1] = _device->attr_GaussReference_read[1];
						fit[2] = _device->attr_GaussReference_read[2];

						double width_ref_uncalibr;
						if(*(_device->attr_WidthCorrectionEnable_read))
						{
							width_ref_uncalibr=sqrt(_device->attr_GaussReference_read[3]*_device->attr_GaussReference_read[3] - _device->widthCorrection*_device->widthCorrection);
						}
						else
						{
							width_ref_uncalibr=_device->attr_GaussReference_read[3];
						}
						fit[3] = width_ref_uncalibr*SIGMA_TO_FWHM_FACTOR;		//TODO: user is aware of using fwhm instead of sigma?
						fit[4] = fit[3];
						DEBUG_STREAM << __func__<<": initial guess from attr_GaussReference_read"<<endl;
					}
					//otherwise start with some default values
					else
					{
						//fit[0] = _device->noise_level; //TODO:
//#ifdef _NON_AVERAGE
//						fit[0] =  //TODO:
//#else
						//average on 4 points: (2 at beginning + 2 at end)/4
						fit[0] = (_device->pLiveData_avg[2 + _device->AnalysisRegionStartPix] +
							_device->pLiveData_avg[5 + _device->AnalysisRegionStartPix] +
							_device->pLiveData_avg[_device->AnalysisRegionEndPix -3] +
							_device->pLiveData_avg[_device->AnalysisRegionEndPix -6])/4;
//#endif
						//try the range 50...10000 as amplitude
						fit[1] = (_device->attr_FWHM_read[1] > 0 && *(_device->attr_FWHMEnable_read)) ? _device->attr_FWHM_read[1] : 50 + (10*counter % 950);
						//if don't have a value from FWHM, use as center the middle of analysis region
						int analysis_region_size = _device->AnalysisRegionEndPix - _device->AnalysisRegionStartPix + 1;
						int center_ind=(int)(((double)(analysis_region_size + ((-1 + 2*(counter%2)) * counter%(analysis_region_size/5))))/2);
						fit[2] = (_device->attr_FWHM_read[0] > 0 && *(_device->attr_FWHMEnable_read)) ? _device->attr_FWHM_read[0] : _device->x_values[center_ind];
						//if don't have a value from FWHM, use as rms a fraction of analysis region
						double width = (_device->x_values[_device->AnalysisRegionEndPix] - _device->x_values[_device->AnalysisRegionStartPix]) / (10.0 + ((1 - 2*(counter%2)) * counter%9));
						fit[3] = (_device->attr_FWHM_read[2] > 0 && *(_device->attr_FWHMEnable_read)) ? _device->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR : width;
						fit[4] = (_device->attr_FWHM_read[2] > 0 && *(_device->attr_FWHMEnable_read)) ? _device->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR : width;
						DEBUG_STREAM << __func__<<": initial guess GENERIC center_ind="<<center_ind<<endl;
					}
					fit[5] = 2;
					fit[6] = 2;
					//DEBUG_STREAM << "readthread::run(): gaussfit initial guess="<<fit[0]<<","<<fit[1]<<","<<fit[2]<<","<<fit[3]<<","<<endl;
					double *data = new double[_device->AnalysisRegionEndPix - _device->AnalysisRegionStartPix + 1];
					for(unsigned int i=_device->AnalysisRegionStartPix; i <= _device->AnalysisRegionEndPix; i++)
					{
						data[i-_device->AnalysisRegionStartPix] = _device->pLiveData_avg[i];
					}
					//size_t fitting_iter;
					unsigned int fitting_iter;
					//SYSTEMTIME ts_end, ts_start;
					//GetSystemTime(&ts_start);
//					ret_fun = (_device->m_gaussfit)->perform(data,_device->AnalysisRegionEndPix - _device->AnalysisRegionStartPix + 1, MAX_GAUSS_ITERATION, fitting_iter, &(_device->x_values[_device->AnalysisRegionStartPix]), fit);
					double tmperr[CONFITEOR_FIT_NUM_PARAM], tmpchi;
					double mean, rms;
					int ret_fun2;
					timespec start, now;
					clock_gettime(CLOCK_MONOTONIC,&start);
					switch(*(_device->attr_FitType_read))
					{
						case GAUSS_FIT:
						default:
							ret_fun = (_device->m_gaussfit)->perform(data,&(_device->x_values[_device->AnalysisRegionStartPix]), _device->AnalysisRegionEndPix - _device->AnalysisRegionStartPix + 1, MAX_GAUSS_ITERATION, 0, fitting_iter, NULL/*weights*/, fit, tmperr, tmpchi);
							ret_fun2 = (_device->m_gaussfit)->compute_rms(fit, &mean, &rms);
							break;
						case GAUSS_FIT_ASYMM:
							ret_fun = (_device->m_asyngaussfit)->perform(data,&(_device->x_values[_device->AnalysisRegionStartPix]), _device->AnalysisRegionEndPix - _device->AnalysisRegionStartPix + 1, MAX_GAUSS_ITERATION, 0, fitting_iter, NULL/*weights*/, fit, tmperr, tmpchi);
							ret_fun2 = (_device->m_asyngaussfit)->compute_rms(fit, &mean, &rms);
							break;
						case CONFITEOR_FIT:
							ret_fun = (_device->m_confiteorfit)->perform(data,&(_device->x_values[_device->AnalysisRegionStartPix]), _device->AnalysisRegionEndPix - _device->AnalysisRegionStartPix + 1, MAX_CONFITEOR_ITERATION, 0, fitting_iter, NULL/*weights*/, fit, tmperr, tmpchi);
							ret_fun2 = (_device->m_confiteorfit)->compute_rms(fit, &mean, &rms);
							break;
					}
					delete [] data;
					clock_gettime(CLOCK_MONOTONIC,&now);
					//printf ("gauss_f: f[0]=%f f[50]=%f\n",f[0],f[50]);
					double fit_time =(now.tv_sec - start.tv_sec)*1000000 + ((double)(now.tv_nsec - start.tv_nsec))/1000;
					fit_counter++;
					if(fit_time < min_fit_time)
						min_fit_time = fit_time;
					if(fit_time > max_fit_time)
						max_fit_time = fit_time;
					tot_fit_time += fit_time;
					//if(fit_counter > 1)
						avg_fit_time = tot_fit_time/(fit_counter);
					//GetSystemTime(&ts_end);
					//DEBUG_STREAM << "readthread::run(): gaussfit performing time: start="<<ts_start.wSecond<<"."<<ts_start.wMilliseconds<<" end="<<ts_end.wSecond<<"."<<ts_end.wMilliseconds<<endl;
					//if(ret_fun)
					bool fit_valid = false;
					//TODO: adjust rms thresh with fwhm
					//eg: if(*(_device->attr_FWHMEnable_read))
					//rms_thresh=fabs(_device->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR/2)
					//if(ret_fun >= 0 && fabs(rms) > 0.08)
					if(0)
					{
						use_last_fit = true;
						_device->attr_GaussFit_read[0] = fit[0];
						_device->attr_GaussFit_read[1] = fit[1];
						//_device->attr_GaussFit_read[2] = fit[2];
						_device->attr_GaussFit_read[2] = mean;
						//double temp = fit[3]/(double)(SIGMA_TO_FWHM_FACTOR);
						double temp = rms/(double)(SIGMA_TO_FWHM_FACTOR);
						if(*(_device->attr_WidthCorrectionEnable_read))
						{
							_device->attr_GaussFit_read[3]=sqrt(temp*temp + _device->widthCorrection*_device->widthCorrection);
						}
						else
						{
							_device->attr_GaussFit_read[3] = (temp > 0) ? temp : -temp;
						}
					}
					else
					{
						use_last_fit = false;
					}
					if(!ret_fun && !ret_fun2)
					{
						//if constraints amp > 0, mean inside analysis region and rms > 1
						if(fit[1] > 0 && mean > _device->x_values[_device->AnalysisRegionStartPix] && mean < _device->x_values[_device->AnalysisRegionEndPix] && fabs(rms) > 0.08)
							fit_valid = true;
					}
					if(fit_valid)
					{
						_device->attr_GaussFit_read[0] = fit[0];
						_device->attr_GaussFit_read[1] = fit[1];
						//_device->attr_GaussFit_read[2] = fit[2];
						_device->attr_GaussFit_read[2] = mean;
						//double temp = fit[3]/(double)(SIGMA_TO_FWHM_FACTOR);
						double temp = rms/(double)(SIGMA_TO_FWHM_FACTOR);
						if(*(_device->attr_WidthCorrectionEnable_read))
						{
							_device->attr_GaussFit_read[3]=sqrt(temp*temp + _device->widthCorrection*_device->widthCorrection);
						}
						else
						{
							_device->attr_GaussFit_read[3] = (temp > 0) ? temp : -temp;
						}
						*(_device->attr_GaussFitCenter_read) = _device->attr_GaussFit_read[2];
						*(_device->attr_GaussFitWidth_read) = _device->attr_GaussFit_read[3];
						double val_buff;
						val_buff = *(_device->attr_GaussFitCenter_read);
						_device->tbuf[BUF_ID_GAUSS_CENTER].push_data(&val_buff,buf_time,buf_bunch_num);
						val_buff = *(_device->attr_GaussFitWidth_read);
						_device->tbuf[BUF_ID_GAUSS_WIDTH].push_data(&val_buff,buf_time,buf_bunch_num);
						_device->GaussFit_valid = true;
						if(!(*(_device->attr_FWHMEnable_read)) || !_device->preferFWHM)
							*(_device->attr_Center_read) = _device->attr_GaussFit_read[2];
							*(_device->attr_Width_read) = _device->attr_GaussFit_read[3];
						DEBUG_STREAM << "readthread::run(): FIT VALID! fit_time="<<fit_time << "["<<min_fit_time<<","<<max_fit_time<<"]->"<<avg_fit_time << "    fit[1]="<<fit[1]<<" mean="<<mean<<" rms="<<rms;
					}
					else
					{
						DEBUG_STREAM << "readthread::run(): FIT NOT VALID: ret_fun="<<ret_fun<<" ret_fun2="<<ret_fun2<<" fit[1]="<<fit[1]<<" mean="<<mean<<" rms="<<rms<<" fit_time="<<fit_time << "["<<min_fit_time<<","<<max_fit_time<<"]->"<<avg_fit_time;
						_device->GaussFit_valid = false;
					}

				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::::::::::  AREA processing :::::::::::::::::::::::
				//________________________________________________________________
				if(*(_device->attr_AreaEnable_read))
				{
					size_t x_start, x_end;
					double wl_start, wl_end;
					bool region_ok = false;
					if((*(_device->attr_FWHMEnable_read) || (*(_device->attr_GaussFitEnable_read) && _device->GaussFit_valid)) &&
						*(_device->attr_Center_read) > _device->attr_RegionOfInterest_read[0] &&
						*(_device->attr_Center_read) < _device->attr_RegionOfInterest_read[1] && !_device->areaInAnalysisRegion)
					{
						wl_start = *(_device->attr_Center_read) - (*(_device->attr_AreaRegion_read)/2);
						if(wl_start < _device->attr_RegionOfInterest_read[0])
							wl_start = _device->attr_RegionOfInterest_read[0];
						wl_end = *(_device->attr_Center_read) + (*(_device->attr_AreaRegion_read)/2);
						if(wl_end > _device->attr_RegionOfInterest_read[1])
							wl_end = _device->attr_RegionOfInterest_read[1];
						x_start = _device->WaveLenght2Pix(wl_start);
						x_end = _device->WaveLenght2Pix(wl_end);
						region_ok = true;
					}
					else if(_device->attr_AnalysisRegion_read[0] >= _device->attr_RegionOfInterest_read[0] &&
						_device->attr_AnalysisRegion_read[1] <= _device->attr_RegionOfInterest_read[1] && _device->areaInAnalysisRegion)
					{
						wl_start = _device->attr_AnalysisRegion_read[0];
						wl_end = _device->attr_AnalysisRegion_read[1];
						x_start = _device->WaveLenght2Pix(wl_start);
						x_end = _device->WaveLenght2Pix(wl_end);
						region_ok = true;
					}

					if(region_ok)
					{
						double area=0;
						for(size_t x = x_start; x <= x_end; x++)
							area += fabs(_device->pLiveData_avg[x]);
						*(_device->attr_Area_read) = area;
						_device->tbuf[BUF_ID_AREA].push_data(&area,buf_time,buf_bunch_num);
					}
				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::: End of Single Mode Acquirement ::::::::::::::::
				//________________________________________________________________
				if(_device->measure_mode)
				{
					//mutex_sleep->lock();
					pausesec = 1;
					pausenano = 0;
					//mutex_sleep->unlock();
					_device->init_measure_mode = false;
					_device->measure_mode = false;
					_device->set_state(Tango::ON);
					_device->set_status("Data ready");
				}
				/*ret= _device->m_usbCCD.StartMeasuring(&(pNoiseData[_device->noise_frames_counter * _device->m_usbCCD.m_nNumReadOuts * _device->numPixel]));
				if(!ret)
				{
					INFO_STREAM << "readthread::run(): error in StartMeasuring()" << endl;
				}*/

				if(_device->stop_mode)
				{
					_device->live_mode = false;
					_device->stop_mode = false;
					pausesec = 1;
					pausenano = 0;
					StopMeas();
				}
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		//++++++++++++++++++++++++ACQUIRE NOISE MODE+++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		if(_device->acquire_noise_mode || _device->stop_mode)
		{
			//DWORD stat;
			//_device->m_usbCCD.GetMeasStatus(&stat);
			//if(stat == STATUS_DATA_READY)

				//_device->m_usbCCD.GetData(&(pNoiseData[_device->noise_frames_counter * _device->m_usbCCD.m_nNumReadOuts * _device->numPixel]));
				if(counter==0 || (*(_device->attr_SyncMode_read) != 0))
					flag = PrepareMeas();
				else
					flag = 0;
				if(flag==0)
					flag = Measure(&(pNoiseData[_device->noise_frames_counter * _device->numPixel]));
				if(flag < 0)
				{
					if(flag==-1)
					{
						_device->set_state(Tango::FAULT);
						//_device->set_status("Communication error");
						//_device->disconnect_device();
						_device->init_acquire_noise_mode = false;
						_device->acquire_noise_mode = false;
						_device->stop_mode = false;
						pausesec = 1;
						pausenano = 0;
						_device->conn_status = false;
					}
					else if(flag==-2) 
					{
						_device->set_state(Tango::OFF);
					}
					continue;
				}
				else if(flag > 0)
				{
					if(_device->stop_mode)
					{
						_device->acquire_noise_mode=false;
					}
					else
					{
						_device->set_state(Tango::ALARM);
					}
					//_device->set_status("Trigger Missing");
					//_device->disconnect_device();
					continue;
				}
				else
				{
					if(_device->get_state()==Tango::ALARM || _device->get_state()==Tango::OFF)
					{
						_device->set_state(Tango::RUNNING);
						_device->set_status("Acquiring noise");
					}
				}
				_device->noise_frames_counter++;
				DEBUG_STREAM << "readthread::run() DATA_READY!! noise_frames_counter="<<_device->noise_frames_counter <<" noise_frames="<<_device->noise_frames << endl;
				if(_device->noise_frames_counter >= _device->noise_frames)
				{
					double noise_avg=0;
					double min_val=0, max_val=10000;
					int tmp_count=0;
					//for every pixel
					for(unsigned int j=0; j < _device->noise_frames * _device->numPixel; j++)
					{
						if(pNoiseData[j] >= min_val && pNoiseData[j] <= max_val)
						{
							noise_avg += pNoiseData[j];
							tmp_count++;
						}
					}
					DEBUG_STREAM << "readthread::run() DATA_READY!! noise sum="<<noise_avg <<" noise pixel counter="<<tmp_count<<" noise pixel total=" << _device->noise_frames * _device->numPixel<< endl;
					//divide the sum per the number of frames
					noise_avg = noise_avg/((double)tmp_count);
					//noise_avg = noise_avg/((double)_device->noise_frames * _device->m_usbCCD.m_nNumReadOuts * _device->numPixel);
					_device->noise_level = (int)noise_avg; //truncate
					*(_device->attr_NoiseLevel_read) = _device->noise_level;

					for(size_t k=0; k<_device->numPixel; k++)
					{
						noise_avg=0;
						for(size_t j=0; j<_device->noise_frames; j++)
						{
							noise_avg += pNoiseData[j* _device->numPixel + k];
						}
						_device->pNoiseMask_avg[k] = noise_avg/_device->noise_frames;
					}

					ofstream  fd;		//startup file
					fd.open(_device->noiseProfileFile.c_str());
					if (!fd)
					{
						stringstream tmp;
						tmp << "Cannot open Noise Profile default file " << _device->noiseProfileFile<< "  err:" << strerror(errno) << endl;
						INFO_STREAM << "readthread::acquire_noise(): " << tmp.str();
				   	}
					else
					{
						DEBUG_STREAM << "readthread::acquire_noise(): file " << _device->noiseProfileFile << " opened successfully" << endl;
						for(size_t i=0; i<_device->numPixel; i++)
						{
							fd << _device->pNoiseMask_avg[i] << endl;
						}
						fd.close();
					}

					try
					{
						Tango::MultiAttribute *attr_list;
						attr_list = _device->get_device_attr();
						Tango::WAttribute &attr_l = attr_list->get_w_attr_by_name ("NoiseLevel");
						attr_l.set_write_value(*(_device->attr_NoiseLevel_read));
					} catch(...)
					{
						DEBUG_STREAM << "readthread::run(): failed setting write value of 'NoiseLevel'";
					}
					DEBUG_STREAM << "readthread::run(): calculated noise average="<<noise_avg<<" -> noise_level="<<_device->noise_level<<endl;

					if(pNoiseData)
						delete [] pNoiseData;
					pNoiseData=NULL;

					//mutex_sleep->lock();
					pausesec = 1;
					pausenano = 0;
					//mutex_sleep->unlock();
					_device->init_acquire_noise_mode = false;
					_device->acquire_noise_mode = false;
					_device->set_state(Tango::ON);
					_device->set_status("Data ready");
				}
				/*else
				{
					ret= _device->m_usbCCD.StartMeasuring(&(pNoiseData[_device->noise_frames_counter * _device->m_usbCCD.m_nNumReadOuts * _device->numPixel]));
					if(!ret)
					{
						INFO_STREAM << "readthread::run(): error in StartMeasuring()" << endl;
					}
				}*/
				if(_device->stop_mode)
				{
					if(pNoiseData)
						delete [] pNoiseData;
					pNoiseData=NULL;
					_device->acquire_noise_mode = false;
					_device->stop_mode = false;
					pausesec = 1;
					pausenano = 0;
					StopMeas();
				}
		}
		if(!(_device->init_live_mode || _device->init_measure_mode || _device->init_acquire_noise_mode)
			&& !(_device->live_mode || _device->measure_mode || _device->stop_mode)
			&& !(_device->acquire_noise_mode || _device->stop_mode))
		{
			try {
				Tango::DevState ccd_state;
				_device->dpmutex->lock();
				_device->dp->read_attribute("State")>>ccd_state;
				_device->dpmutex->unlock();
				if(ccd_state != Tango::ON)
				{
					_device->set_state(Tango::OFF);
				}
			}
			catch(Tango::DevFailed &e)
			{
				_device->dpmutex->unlock();
				INFO_STREAM << __func__ << ": error reading =" << e.errors[0].desc;
				_device->set_state(Tango::OFF);
				_device->set_status(string(e.errors[0].desc));
			}
			
		}
		if(((_device->measure_mode || _device->live_mode || _device->acquire_noise_mode) && ((counter % 100) == 0)) ||
			(!_device->measure_mode && !_device->live_mode && !_device->acquire_noise_mode))
		{
			//DEBUG_STREAM << "readthread::run(): measure_mode=" << _device->measure_mode << " _device->live_mode=" << _device->live_mode << endl;
			//DEBUG_STREAM << "readthread::run(): pausesec=" << pausesec << " pausenano=" << pausenano << endl;
			//DEBUG_STREAM << "readthread::run(): before GetParameters()" << endl;
		}
		abort_sleep(pausesec + ((double)pausenano)/1000000000);
		counter++;
		//if stopped and connection OK -> heartbeat
		if(_device->conn_status && (!_device->measure_mode && !_device->live_mode && !_device->acquire_noise_mode))
		{
			/*int ret = Heartbeat();
			if(ret < 0)
			{
				_device->set_state(Tango::FAULT);
				_device->disconnect_device();
				_device->conn_status = false;
			}*/
		}
		//if stopped and connection NOK -> reconnect
		else if(!_device->measure_mode && !_device->live_mode && !_device->acquire_noise_mode)
		{
			//_device->conn_status = _device->OpenComm(true);
			if(_device->autoStart && _device->conn_status)
			{
				try
				{
					_device->start();
				}
				catch(Tango::DevFailed &e)
				{
					INFO_STREAM << __func__ << ": error autostarting=" << e.errors[0].desc;
				}
			}
		}

	}	//end while(!abortflag)
	if(pNoiseData)
		delete [] pNoiseData;
	pNoiseData=NULL;
	DEBUG_STREAM << "readthread::run(): exit!!!" << endl;
}

//+------------------------------------------------------------------
//
//	method:			readthread::abort_sleep
//
//	description:	Resume from sleep if abort_flag set (sec.)
//
//-------------------------------------------------------------------
void readthread::abort_sleep(double time)
{
	omni_mutex_lock sync(*this);
	long time_millis = (long)(time*1000);
	int res = wait(time_millis);
}

int readthread::PrepareMeas()
{
	CCDSpectrometer *_device = static_cast<CCDSpectrometer *>(device);
	Tango::DevState ccd_state;
	int retry_cnt=0;
	do
	{
		retry_cnt++;
		try {
			Tango::DevState data;
			_device->dpmutex->lock();
			_device->dp->read_attribute("State")>>ccd_state;
			_device->dpmutex->unlock();
		}
		catch(Tango::DevFailed &e)
		{
			_device->dpmutex->unlock();
			INFO_STREAM << __func__ << ": error reading =" << e.errors[0].desc;
			return -1;
		}
		if(ccd_state == Tango::INIT || ccd_state == Tango::OFF)
			abort_sleep(1);
	} while ((ccd_state == Tango::INIT || ccd_state == Tango::OFF) && retry_cnt < 30);
	if(retry_cnt >= 30)
		return -1;
	return 0;
}

int readthread::Measure(double *l_pSpectrum)
{
	CCDSpectrometer *_device = static_cast<CCDSpectrometer *>(device);
	Tango::DevLong camera_counter;
	int retry_cnt=0;
	do
	{
		retry_cnt++;
		try {
			//Tango::DevDouble data;
			_device->dpmutex->lock();
			_device->dp->read_attribute(COUNTER_ATTR)>>camera_counter;
			_device->dpmutex->unlock();
		}
		catch(Tango::DevFailed &e)
		{
			_device->dpmutex->unlock();
			INFO_STREAM << __func__ << ": error reading =" << e.errors[0].desc;
			_device->set_status("Camera Stopped");
			abort_sleep(0.5);
			return -2;
		}
		if(camera_counter == last_camera_counter)
			abort_sleep(0.01);
	} while (camera_counter == last_camera_counter && retry_cnt < 100);
	//DEBUG_STREAM << __func__ << ": camera_counter=" << camera_counter << " retry_cnt="<<retry_cnt;
	last_camera_counter = camera_counter;
	if(retry_cnt >= 100)
	{
		if((*(_device->attr_SyncMode_read) != 0)) //-> HW
		{
			ERROR_STREAM << "Communication error (read Measurement TIMEOUT)";
			if(_device->live_mode || _device->measure_mode || _device->acquire_noise_mode)
				_device->set_status("Timeout reading");
			return 1;
		}
		else
		{
			ERROR_STREAM << "Communication error (read Measurement failed)";
			_device->set_status("Timeout reading");
			return -1;
		}
	}

	try {
		vector<Tango::DevDouble> data;
		Tango::DeviceAttribute dattr;
		_device->dpmutex->lock();
		_device->dp->read_attribute(DATA_ATTR)>>(data);
		for(size_t i=0; i<_device->numPixel; i++)
		{
			if(i < _device->RegionOfInterestStartPix || i > _device->RegionOfInterestEndPix)
			{
				l_pSpectrum[i] = 0;
			}
			else if(i-_device->RegionOfInterestStartPix < data.size())
			{
				l_pSpectrum[i] = data[i-_device->RegionOfInterestStartPix];
				//DEBUG_STREAM << __func__ << ": read data.size()=" << data.size() << " data[data.size()-1]="<<data[data.size()-1];
			}
		}
		_device->dpmutex->unlock();
	}
	catch(Tango::DevFailed &e)
	{
		_device->dpmutex->unlock();
		INFO_STREAM << __func__ << ": error reading =" << e.errors[0].desc;
		if(_device->get_state() == Tango::OFF)
			return -2;
		return -1;
	}
	return 0;
}

void readthread::StopMeas()
{
}

}


